import platform
import boto3
import stat
import json
import os

if platform.system() == 'Windows':
    basedir = os.getenv('HOMEPATH')
else:
    basedir = os.getenv('HOME')

DIRECTORY = os.path.join(basedir, '.aisquared')
AWS_FILE = os.path.join(DIRECTORY, 'aws_config.json')


class AWSClient:
    """
    Client for interacting with and viewing models stored in AWS S3 Storage
    """

    def __init__(self):
        self._client = boto3.client('s3')

    def get_default_bucket(self):
        """
        Returns the name of the default bucket
        """
        try:
            with open(AWS_FILE, 'r') as f:
                loaded = json.load(f)
            return loaded['defaultBucket']
        except Exception:
            raise ValueError(
                'It does not appear that a default bucket is configured. Try running `AWSClient.configure()` with a default bucket name to configure')

    def list_models(self, bucket=None, detailed=False):
        """
        List models in storage

        Parameters
        ----------
        bucket : str or None (default None)
            Bucket to search for models in
        detailed : bool (default False)
            Whether to return additional metadata about models
        """
        if bucket is None:
            bucket = self.get_default_bucket()
        listed = self._client.list_objects_v2(Bucket=bucket)
        if detailed:
            return listed
        try:
            return [content['Key'] for content in listed['Contents']]
        except Exception:
            return listed

    def delete_model(self, model_name, bucket=None):
        """
        Delete a model

        Parameters
        ----------
        model_name : str
            The name of the model to delete
        bucket : str or None (default None)
            Bucket to delete from
        """
        if bucket is None:
            bucket = self.get_default_bucket()
        self._client.delete_object(Bucket=bucket, Key=model_name)

    def download_model(self, model_name, bucket=None):
        """
        Download a model

        Parameters
        ----------
        model_name : str
            The name of the model to download
        bucket : str or None (default None)
            Bucket to download from
        """
        if bucket is None:
            bucket = self.get_default_bucket()
        self._client.download_file(bucket, model_name, model_name)

    def upload_model(self, model_path, bucket=None, overwrite=False):
        """
        Upload a model

        Parameters
        ----------
        model_path : str or path-like
            The path of the model saved locally
        bucket : str or None (default None)
            The bucket to upload to
        overwrite : bool (default False)
            Whether to overwrite the model if it already exists
        """
        if bucket is None:
            bucket = self.get_default_bucket()
        object_name = os.path.basename(model_path)
        if os.path.splitext(object_name)[-1] != '.air':
            raise ValueError(
                'It does not appear that the specified file is the correct file type')
        if not overwrite:
            if object_name in self.list_models():
                raise ValueError(
                    'Model with the same name already exists. To overwrite the model, set overwrite to True')
        self._client.upload_file(model_path, bucket, object_name)

    @staticmethod
    def configure(default_bucket):
        """Configure the default behaviors to use"""
        if not os.path.exists(DIRECTORY):
            os.makedirs(DIRECTORY)
        with open(AWS_FILE, 'w') as f:
            json.dump({'defaultBucket': default_bucket}, f)
        os.chmod(AWS_FILE, stat.S_IREAD | stat.S_IWRITE)



from aisquared.base import BaseObject


class RegressionFeedback(BaseObject):
    """
    Feedback for regression
    """

    def __init__(self):
        super().__init__()

    def to_dict(self):
        """
        Return the object as a dictionary
        """
        return {
            'className': 'RegressionFeedback',
            'params': dict()
        }



from aisquared.base import BaseObject


class ReverseMLWorkflow(BaseObject):
    """
    Interaction with a ReverseML CSV stored in S3
    """

    def __init__(
        self,
        bucket,
        filename,
        column,
        period=None
    ):
        """
        Parameters
        ----------
        bucket : str
            The bucket the file lives in
        filename : str
            The name for the specific file
        column : str
            The column name in the CSV file
        period : None or int (default None)
            The period for this to run
        """
        super().__init__()
        self.bucket = bucket
        self.filename = filename
        self.column = column
        self.period = period

    @property
    def bucket(self):
        return self._bucket

    @bucket.setter
    def bucket(self, value):
        if not isinstance(value, str):
            raise TypeError('bucket must be string')
        self._bucket = value

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        if not isinstance(value, str):
            raise TypeError('filename must be string')
        self._filename = value

    @property
    def column(self):
        return self._column

    @column.setter
    def column(self, value):
        if not isinstance(value, str):
            raise TypeError('column must be a string')
        self._column = value

    @property
    def period(self):
        return self._period

    @period.setter
    def period(self, value):
        if value is not None:
            if not isinstance(value, int):
                raise TypeError('period must be int if provided')
            if value < 1:
                raise ValueError('period must be greater than or equal to 1')
        self._period = value

    def to_dict(self):
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ReverseMLWorkflow',
            'params': {
                'bucket': self.bucket,
                'fileName': self.filename,
                'column': self.column,
                'period': self.period
            }
        }



"""
The aisquared.base package contains both some basic objects that are used across the aisquared package backend and some objects which are designed to facilitate simple use cases of the technology.
"""

from .BaseObject import BaseObject
from .CustomObject import CustomObject
from .rendering import LOCATIONS, COLORS, BADGES, WORD_LISTS, QUALIFIERS
from .stages import ALLOWED_STAGES



from .BaseObject import BaseObject


class CustomObject(BaseObject):
    """
    Custom class that allows the user to define custom classes for configuration
    """

    def __init__(self, class_name, **kwargs):
        """
        Parameters
        ----------
        class_name : str
            The name for the class
        **kwargs : additional keyword arguments
            Other parameters to be used
        """
        super().__init__()
        self.class_name = class_name
        self.params = kwargs

    def to_dict(self):
        """
        Get the object as a dictionary
        """
        return {
            'className': self.class_name,
            'params': self.params
        }



import json


class BaseObject:
    """
    Base class used for all other classes within the aisquared package
    """

    def __init__(self):
        pass

    def to_dict(self):
        """
        Get the object as a dictionary
        """
        raise NotImplementedError

    def to_json(self):
        """
        Return the object as a json string
        """
        return json.dumps(self.to_dict())



"""
The aisquared.config.rendering subpackage contains objects for configuring how rendering of predictions is to occur.
"""

from .ImageRendering import ImageRendering
from .ObjectRendering import ObjectRendering
from .DocumentRendering import DocumentRendering
from .WordRendering import WordRendering
from .FilterRendering import FilterRendering



from aisquared.base import BaseObject


class DeployedModel(BaseObject):
    """
    Interaction with a remote model
    """

    def __init__(
        self,
        url,
        input_type,
        secret='request',
        header=None
    ):
        """
        Parameters
        ----------
        url : str
            The base URL for the remote endpoint
        input_type : str
            The input type to the model. Either one of 'cv', 'text', or 'tabular'
        secret : str (default 'request')
            The secret key used to interact with the service. Default value of 'request'
            indicates that the user inputs the key whenever the analytic is started again
        header : dict or None (default None)
            Header to use when calling the endpoint
        """
        super().__init__()
        self.url = url
        self.input_type = input_type
        self.secret = secret
        self.header = header

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def input_type(self):
        return self._input_type

    @input_type.setter
    def input_type(self, value):
        self._input_type = value

    @property
    def secret(self):
        return self._secret

    @secret.setter
    def secret(self, value):
        self._secret = value

    @property
    def header(self):
        return self._header

    @header.setter
    def header(self, value):
        self._header = value

    def to_dict(self):
        """
        Get the config object as a dictionary
        """
        return {
            'className': 'DeployedModel',
            'params': {
                'url': self.url,
                'inputType': self.input_type,
                'secret': self.secret,
                'header': self.header
            }
        }



"""
The aisquared.config.analytic subpackage contains objects for packaging individual analytics.
"""

from .DeployedAnalytic import DeployedAnalytic
from .DeployedModel import DeployedModel
from .LocalModel import LocalModel
from .LocalAnalytic import LocalAnalytic
from .ReverseMLWorkflow import ReverseMLWorkflow



from aisquared.base import BaseObject, CustomObject, ALLOWED_STAGES
from aisquared.config.harvesting import ImageHarvester, TextHarvester, InputHarvester
from aisquared.config.preprocessing.tabular import TabularPreprocessor
from aisquared.config.preprocessing.image import ImagePreprocessor
from aisquared.config.preprocessing.text import TextPreprocessor
from aisquared.config.analytic import DeployedAnalytic, DeployedModel, LocalModel, LocalAnalytic, ReverseMLWorkflow
from aisquared.config.postprocessing import BinaryClassification, MulticlassClassification, ObjectDetection, Regression
from aisquared.config.rendering import ImageRendering, ObjectRendering, DocumentRendering, WordRendering, FilterRendering
from aisquared.config.feedback import SimpleFeedback, BinaryFeedback, MulticlassFeedback, RegressionFeedback, ModelFeedback, QualitativeFeedback

import tensorflowjs as tfjs
import tensorflow as tf
import shutil
import json
import os

HARVESTING_CLASSES = (
    ImageHarvester,
    TextHarvester,
    InputHarvester,
    CustomObject
)

PREPROCESSING_CLASSES = (
    TabularPreprocessor,
    ImagePreprocessor,
    TextPreprocessor,
    CustomObject
)

ANALYTIC_CLASSES = (
    DeployedAnalytic,
    DeployedModel,
    LocalModel,
    LocalAnalytic,
    ReverseMLWorkflow,
    CustomObject
)

POSTPROCESSING_CLASSES = (
    BinaryClassification,
    MulticlassClassification,
    ObjectDetection,
    Regression,
    CustomObject
)

RENDERING_CLASSES = (
    ObjectRendering,
    ImageRendering,
    DocumentRendering,
    WordRendering,
    FilterRendering,
    CustomObject
)

FEEDBACK_CLASSES = (
    ModelFeedback,
    SimpleFeedback,
    BinaryFeedback,
    MulticlassFeedback,
    RegressionFeedback,
    QualitativeFeedback
)

LOCAL_CLASSES = (
    LocalModel,
    LocalAnalytic,
    CustomObject
)


class ModelConfiguration(BaseObject):
    """
    Configuration object for deploying a model or analytic
    """

    def __init__(
            self,
            name,
            harvesting_steps,
            preprocessing_steps,
            analytic,
            postprocessing_steps,
            rendering_steps,
            feedback_steps=None,
            stage=ALLOWED_STAGES[0],
            version=None,
            description='',
            mlflow_uri=None,
            mlflow_user=None,
            mlflow_token=None,
            owner=None,
            url='*',
            auto_run=False
    ):
        """
        Parameters
        ----------
        name : str
            The name of the deployed analytic
        harvesting_steps : None, Harvesting object or list of Harvesting objects
            Harvesters to use with the analytic
        preprocessing_steps : Preprocessing object or list of Preprocessing objects or None
            Preprocessers to use
        analytic : Analytic object or list of Analytic objects
            Analytics to use
        postprocessing_steps : Postprocessing object or list of Postprocessing objects or None
            Postprocessers to use
        rendering_steps : Rendering object or list of Rendering objects or None
            Renderers to use
        feedback_steps : None, Feedback object or list of Feedback objects or None (default None)
            Feedback steps to use
        stage : str (default 'experimental')
            The stage of the model, from 'experimental', 'staging', 'production'
        version : str or None (default None)
            Version of the analytic
        description : str (default '')
            The description of the analytic
        mlflow_uri : str or None (default None)
            MLFlow URI to use, if applicable
        mlflow_user : str or None (default None)
            MLFlow user to use, if applicable
        mlflow_token : str or None (default None)
            MLFlow token to use, if applicable
        owner : str or None (default None)
            The owner of the model
        url : str (default '*')
            URL or URL pattern to match
        auto_run : bool (default False)
            Whether to automatically run this file when on a valid page
        """
        super().__init__()
        self.name = name
        self.harvesting_steps = harvesting_steps
        self.preprocessing_steps = preprocessing_steps
        self.analytic = analytic
        self.postprocessing_steps = postprocessing_steps
        self.rendering_steps = rendering_steps
        self.stage = stage
        self.feedback_steps = feedback_steps
        self.version = version
        self.description = description
        self.mlflow_uri = mlflow_uri
        self.mlflow_user = mlflow_user
        self.mlflow_token = mlflow_token
        self.owner = owner
        self.url = url
        self.auto_run = auto_run

    # name
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = str(value)

    # harvesting_steps
    @property
    def harvesting_steps(self):
        return self._harvesting_steps

    @harvesting_steps.setter
    def harvesting_steps(self, value):
        harvesting_classes = HARVESTING_CLASSES + (ModelConfiguration,)
        if value is None or (isinstance(value, list) and all([val is None for val in value])):
            self._harvesting_steps = value
        elif isinstance(value, harvesting_classes):
            self._harvesting_steps = [value]
        elif isinstance(value, list) and all([isinstance(val, harvesting_classes) for val in value]):
            self._harvesting_steps = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, harvesting_classes) for val in value for v in val]):
            self._harvesting_steps = value
        else:
            raise ValueError(
                'harvesting_steps must be a None, single Harvester object, a list of Harvester objects, or a list of list of harvester objects')

    # preprocessing_steps
    @property
    def preprocessing_steps(self):
        return self._preprocessing_steps

    @preprocessing_steps.setter
    def preprocessing_steps(self, value):
        if value is None:
            self._preprocessing_steps = value
        elif isinstance(value, PREPROCESSING_CLASSES):
            self._preprocessing_steps = [value]
        elif isinstance(value, list) and all([isinstance(val, PREPROCESSING_CLASSES) for val in value]):
            self._preprocessing_steps = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, PREPROCESSING_CLASSES) for val in value for v in val]):
            self._preprocessing_steps = value
        elif value is None:
            self._preprocessing_steps = value
        else:
            raise ValueError(
                'preprocessing_steps must a single Preprocessor object, a list of Preprocessor objects, or a list of list of preprocessor objects')

    # analytic
    @property
    def analytic(self):
        return self._analytic

    @analytic.setter
    def analytic(self, value):
        if isinstance(value, ANALYTIC_CLASSES):
            self._analytic = [value]
        elif isinstance(value, list) and all([isinstance(val, ANALYTIC_CLASSES) for val in value]):
            self._analytic = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, ANALYTIC_CLASSES) for val in value for v in val]):
            self._analytic = value
        else:
            raise ValueError(
                'analytic must be a single Analytic object, a list of Analytic objects, or a list of list of Analtyic objects')

    # postprocessing_steps
    @property
    def postprocessing_steps(self):
        return self._postprocessing_steps

    @postprocessing_steps.setter
    def postprocessing_steps(self, value):
        if value is None:
            self._postprocessing_steps = value
        elif isinstance(value, POSTPROCESSING_CLASSES):
            self._postprocessing_steps = [value]
        elif isinstance(value, list) and all([isinstance(val, POSTPROCESSING_CLASSES) for val in value]):
            self._postprocessing_steps = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, POSTPROCESSING_CLASSES) for val in value for v in val]):
            self._postprocessing_steps = value
        elif value is None:
            self._postprocessing_steps = value
        else:
            raise ValueError(
                'postprocessing_steps must be a single Postprocessing object, a list of Postprocessing objects, or a list of list of Postprocessing objects')

    # rendering_steps
    @property
    def rendering_steps(self):
        return self._rendering_steps

    @rendering_steps.setter
    def rendering_steps(self, value):
        if isinstance(value, RENDERING_CLASSES) or value is None:
            self._rendering_steps = [value]
        elif isinstance(value, list) and all([isinstance(val, RENDERING_CLASSES) for val in value]):
            self._rendering_steps = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, RENDERING_CLASSES) for val in value for v in val]):
            self._rendering_steps = value
        else:
            raise ValueError(
                'rendering_steps must be a single Rendering object, a list of Rendering objects, a list of list of Rendering objects, or None')

    # feedback_steps
    @property
    def feedback_steps(self):
        return self._feedback_steps

    @feedback_steps.setter
    def feedback_steps(self, value):
        if value is None:
            self._feedback_steps = value
        elif isinstance(value, FEEDBACK_CLASSES):
            self._feedback_steps = [value]
        elif isinstance(value, list) and all([isinstance(val, FEEDBACK_CLASSES) for val in value]):
            self._feedback_steps = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, FEEDBACK_CLASSES) for val in value for v in val]):
            self._feedback_steps = value
        else:
            raise ValueError(
                'feedback_steps must be a single Feedback object, a list of Feedback objects, a list of list of Feedback objects, or None')

    # stage
    @property
    def stage(self):
        return self._stage

    @stage.setter
    def stage(self, value):
        if value not in ALLOWED_STAGES:
            raise ValueError(f'stage must be one of {ALLOWED_STAGES}')
        self._stage = value

    # version
    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        self._version = str(value) if value is not None else value

    # description
    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = str(value)

    # mlflow_uri
    @property
    def mlflow_uri(self):
        return self._mlflow_uri

    @mlflow_uri.setter
    def mlflow_uri(self, value):
        self._mlflow_uri = value

    # mlflow_user
    @property
    def mlflow_user(self):
        return self._mlflow_user

    @mlflow_user.setter
    def mlflow_user(self, value):
        self._mlflow_user = value

    # mlflow_token
    @property
    def mlflow_token(self):
        return self._mlflow_token

    @mlflow_token.setter
    def mlflow_token(self, value):
        self._mlflow_token = value

    # owner
    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        if not isinstance(value, str) and value is not None:
            raise ValueError('owner must be a string or None')
        self._owner = value

    # url
    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        if not isinstance(value, str):
            raise ValueError('url must be string')
        self._url = value

    # auto_run
    @property
    def auto_run(self):
        return self._auto_run

    @auto_run.setter
    def auto_run(self, value):
        if not isinstance(value, bool):
            raise TypeError('auto_run must be Boolean valued')
        self._auto_run = value

    # harvester_dict
    @property
    def harvester_dict(self):
        harvesting_classes = HARVESTING_CLASSES + (ModelConfiguration,)
        if self.harvesting_steps is None or (isinstance(self.harvesting_steps, list) and all([val is None for val in self.harvesting_steps])):
            return None
        elif isinstance(self.harvesting_steps, list) and all([isinstance(val, harvesting_classes) for val in self.harvesting_steps]):
            return [val.to_dict() for val in self.harvesting_steps]
        else:
            return [
                [v.to_dict() for v in val] for val in self.harvesting_steps
            ]

    # preprocessing_dict
    @property
    def preprocesser_dict(self):
        if self.preprocessing_steps is None:
            return self.preprocessing_steps
        elif isinstance(self.preprocessing_steps, list) and all([isinstance(val, PREPROCESSING_CLASSES) for val in self.preprocessing_steps]):
            return [val.to_dict() for val in self.preprocessing_steps]
        else:
            return [
                [v.to_dict() for v in val] for val in self.preprocessing_steps
            ]

    # analytic dict
    @property
    def analytic_dict(self):
        if isinstance(self.analytic, list) and all([isinstance(val, ANALYTIC_CLASSES) for val in self.analytic]):
            return [val.to_dict() for val in self.analytic]
        else:
            return [
                [v.to_dict() for v in val] for val in self.analytic
            ]

    # postprocesser_dict
    @property
    def postprocesser_dict(self):
        if self.postprocessing_steps is None:
            return self.postprocessing_steps
        elif isinstance(self.postprocessing_steps, list) and all([isinstance(val, POSTPROCESSING_CLASSES) for val in self.postprocessing_steps]):
            return [val.to_dict() for val in self.postprocessing_steps]
        else:
            return [
                [v.to_dict() for v in val] for val in self.postprocessing_steps
            ]

    # render_dict
    @property
    def render_dict(self):
        if self.rendering_steps[0] is None:
            return []
        elif isinstance(self.rendering_steps, list) and all([isinstance(val, RENDERING_CLASSES) for val in self.rendering_steps]):
            return [val.to_dict() for val in self.rendering_steps]
        else:
            return [
                [v.to_dict() for v in val] for val in self.rendering_steps
            ]

    # feedback_dict
    @property
    def feedback_dict(self):
        if self.feedback_steps is None:
            return self.feedback_steps
        elif isinstance(self.feedback_steps, list) and all([isinstance(val, FEEDBACK_CLASSES) for val in self.feedback_steps]):
            return [val.to_dict() for val in self.feedback_steps]
        else:
            return [
                [v.to_dict() for v in val] for val in self.feedback_steps
            ]

    def get_model_filenames(self):
        """
        Get filenames for all models in the configuration
        """
        filenames = []
        if self.harvesting_steps is None or (isinstance(self.harvesting_steps, list) and all([val is None for val in self.harvesting_steps])):
            harvesting_list = []
        elif isinstance(self.harvesting_steps[0], list):
            harvesting_list = [
                h for harvester in self.harvesting_steps for h in harvester]
        else:
            harvesting_list = self.harvesting_steps
        for harvester in harvesting_list:
            if isinstance(harvester, ModelConfiguration):
                filenames.extend(harvester.get_model_filenames())

        if isinstance(self.analytic[0], ANALYTIC_CLASSES):
            for a in self.analytic:
                if isinstance(a, LOCAL_CLASSES):
                    try:
                        filenames.append(a.path)
                    except:
                        pass
        else:
            for analytic in self.analytic:
                for a in analytic:
                    if isinstance(a, LOCAL_CLASSES):
                        try:
                            filenames.append(a.path)
                        except:
                            pass
        return [f for f in filenames if f is not None]

    def to_dict(self):
        """
        Get the object as a dictionary
        """
        return {
            'className': 'ModelConfiguration',
            'params': {
                'name': self.name,
                'harvestingSteps': self.harvester_dict,
                'preprocessingSteps': self.preprocesser_dict,
                'analytics': self.analytic_dict,
                'postprocessingSteps': self.postprocesser_dict,
                'renderingSteps': self.render_dict,
                'feedbackSteps': self.feedback_dict,
                'stage': self.stage,
                'version': self.version,
                'description': self.description,
                'mlflowUri': self.mlflow_uri,
                'mlflowUser': self.mlflow_user,
                'mlflowToken': self.mlflow_token,
                'owner': self.owner,
                'url': self.url,
                'autoRun': self.auto_run
            }
        }

    def compile(self, filename=None, dtype=None):
        """
        Compile the object into a '.air' file, which can then be dragged and
        dropped into applications using the AI Squared JavaScript SDK

        Parameters
        ----------
        filename : path-like or None (default None)
            Filename to compile to. If None, defaults to '{NAME}.air', where {NAME} is the
            name of the analytic
        dtype : str or None (default None)
            The datatype to use for the model weights. If None, defaults to 'float32'
        """
        if filename is None:
            filename = self.name + '.air'

        if dtype is None:
            dtype_map = None
        else:
            dtype_map = {dtype: '*'}

        dirname = os.path.join('.', os.path.splitext(filename)[0])

        # write the object as json config
        os.makedirs(dirname, exist_ok=True)
        with open(os.path.join(dirname, 'config.json'), 'w') as f:
            json.dump(self.to_dict(), f)

        # go through the files and copy them/make them tfjs files
        filenames = self.get_model_filenames()
        if len(filenames) != 0:
            for f in filenames:
                if os.path.splitext(f)[-1] == '.h5':
                    model = tf.keras.models.load_model(f)
                    model_dir = os.path.join(dirname, os.path.split(f)[-1])
                    tfjs.converters.save_keras_model(
                        model, model_dir, quantization_dtype_map=dtype_map)
                else:
                    shutil.copy(f, dirname)

        # go through the entire directory of dirname, grab all files, and make
        # the archive file
        shutil.make_archive(filename, 'zip', dirname)

        # Move the archive file to just have .air
        shutil.move(filename + '.zip', filename)

        # Remove the temp directory
        shutil.rmtree(dirname, ignore_errors=True)



from azure.storage.blob import BlobServiceClient
from getpass import getpass
import platform
import stat
import json
import os

if platform.system() == 'Windows':
    basedir = os.getenv('HOMEPATH')
else:
    basedir = os.getenv('HOME')

DIRECTORY = os.path.join(basedir, '.aisquared')
AZURE_FILE = os.path.join(DIRECTORY, 'azure_config.json')


class AzureClient:
    """
    Client for interacting with and viewing models stored in Azure
    """

    def __init__(self):
        try:
            with open(AZURE_FILE, 'r') as f:
                loaded = json.load(f)
            connection_string = loaded['connectionString']
            self._client = BlobServiceClient.from_connection_string(
                connection_string)
        except Exception:
            raise RuntimeError(
                'Connection could not be made. Please ensure you have configured the connection string by running `AzureClient.configure`')

    def get_default_container(self):
        """
        Returns the name of the default container
        """
        try:
            with open(AZURE_FILE, 'r') as f:
                loaded = json.load(f)
            return loaded['defaultContainer']
        except Exception:
            raise ValueError(
                'It does not appear that a default container is configured. Try running `AzureClient.configure()` to configure')

    def list_models(self, container=None, detailed=False):
        """
        List models in storage

        Parameters
        ----------
        container : str or None (default None)
            Container to search for models in
        detailed : bool (default False)
            Whether to return additional metadata about models
        """
        if container is None:
            container = self.get_default_container()
        container_client = self._client.get_container_client(container)
        listed = container_client.list_blobs()
        if detailed:
            return list(listed)
        return [
            content['name'] for content in listed
        ]

    def delete_model(self, model_name, container=None):
        """
        Delete a model

        Parameters
        ----------
        model_name : str
            The name of the model to delete
        container : str or None (default None)
            Container to delete from
        """
        if container is None:
            container = self.get_default_container()
        container_client = self._client.get_container_client(container)
        container_client.delete_blob(model_name)

    def download_model(self, model_name, container=None):
        """
        Download a model

        Parameters
        ----------
        model_name : str
            The name of the model to download
        container : str or None (default None)
            Container to download from
        """
        if container is None:
            container = self.get_default_container()
        container_client = self._client.get_container_client(container)
        with open(model_name, 'wb') as f:
            f.write(container_client.download_blob(model_name).readall())

    def upload_model(self, model_path, container=None, overwrite=False):
        """
        Upload a model

        Parameters
        ----------
        model_path : str or path-like
            The path of the model saved locally
        container : str or None (default None)
            The container to upload to
        overwrite : bool (default False)
            Whether to overwrite model if it already exists
        """
        if container is None:
            container = self.get_default_container()
        container_client = self._client.get_container_client(container)
        object_name = os.path.basename(model_path)
        if os.path.splitext(object_name)[-1] != '.air':
            raise ValueError(
                'It does not appear that the specified file is the correct file type')
        with open(object_name, 'rb') as f:
            container_client.upload_blob(
                object_name,
                f,
                overwrite=overwrite
            )

    @staticmethod
    def configure(
        connection_string=None,
        default_container=None
    ):
        """Configure the default behavior to use"""
        if not os.path.exists(DIRECTORY):
            os.makedirs(DIRECTORY)
        if connection_string is None:
            connection_string = getpass('Connection String: ')
        if default_container is None:
            default_container = input('Default Container: ')
        with open(AZURE_FILE, 'w') as f:
            json.dump(
                {
                    'connectionString': connection_string,
                    'defaultContainer': default_container
                },
                f
            )
        os.chmod(AZURE_FILE, stat.S_IREAD | stat.S_IWRITE)



from aisquared.base import BaseObject


class BinaryFeedback(BaseObject):
    """
    Feedback for binary classification
    """

    def __init__(self, label_map):
        super().__init__()
        self.label_map = label_map

    @property
    def label_map(self):
        return self._label_map

    @label_map.setter
    def label_map(self, value):
        if not isinstance(value, list) or len(value) != 2:
            raise ValueError('label_map must be list of length 2')
        self._label_map = value

    def to_dict(self):
        """
        Return the object as a dictionary
        """
        return {
            'className': 'BinaryFeedback',
            'params': {
                'labelMap': self.label_map
            }
        }



from aisquared.serving import deploy_model, get_remote_prediction
from aisquared.remote import AWSClient, AzureClient
from mann.utils import get_custom_objects
import click
import json

_ALLOWED_CLIENTS = ['aws', 'azure']


def _get_client(client_type):

    # Normalize the client type name
    client_type = client_type.lower()

    # Create the client
    if client_type == 'aws':
        client = AWSClient()
    elif client_type == 'azure':
        client = AzureClient()

    return client


@click.group()
@click.pass_context
def aisquared(context):
    pass


@aisquared.command('deploy')
@click.argument('saved-model', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('model-type', type=str)
@click.option('--host', '-h', type=str, default='127.0.0.1')
@click.option('--port', '-p', type=int, default=2244)
def deploy(saved_model, model_type, host, port):
    """
    Deploy a model to a local endpoint
    """
    deploy_model(
        saved_model,
        model_type,
        host,
        port,
        get_custom_objects()
    )


@aisquared.command('predict')
@click.argument('data', type=click.Path(exists=True, file_okay=True, dir_okay=True))
@click.option('--host', '-h', type=str, default='127.0.0.1')
@click.option('--port', '-p', type=int, default=2244)
@click.option('--outfile', '-o', default=None, type=click.Path(exists=False, file_okay=True, dir_okay=False))
def predict(data, host, port, outfile):
    """
    Get predictions from a deployed model
    """
    # Load the json
    with open(data, 'r') as f:
        data = json.load(f)

    # Get the predictions
    predictions = get_remote_prediction(
        data,
        host,
        port
    )

    # Either print the predictions or write to the outfile
    if not outfile:
        print(predictions)
    else:
        with open(outfile, 'w') as f:
            json.dump(predictions, f)


@aisquared.group()
def airfiles():
    """
    Utilities for listing, downloading, uploading, and deleting .air files
    """
    pass


@airfiles.command('list')
@click.argument('client-type', type=click.Choice(_ALLOWED_CLIENTS, case_sensitive=False))
@click.option('--bucket', '-b', type=str, default=None)
@click.option('--detailed/--no-detailed', default=False)
def list_models(client_type, bucket, detailed):
    """
    List .air files in a specific bucket or container
    """

    # Create the client
    client = _get_client(client_type)

    # Print out the response
    models = client.list_models(
        bucket,
        detailed=detailed
    )
    if not detailed:
        print('\n'.join(models))
    else:
        print(models)


@airfiles.command('delete')
@click.argument('client-type', type=click.Choice(_ALLOWED_CLIENTS, case_sensitive=False))
@click.argument('model-name', type=str)
@click.option('--bucket', '-b', type=str, default=None)
def delete(client_type, model_name, bucket):
    """
    Delete an .air file
    """

    # Create the client
    client = _get_client(client_type)

    # Delete the model
    client.delete_model(
        model_name,
        bucket
    )


@airfiles.command('download')
@click.argument('client-type', type=click.Choice(_ALLOWED_CLIENTS, case_sensitive=False))
@click.argument('model-name', type=str)
@click.option('--bucket', '-b', type=str, default=None)
def download(client_type, model_name, bucket):
    """
    Download an .air file
    """

    # Create the client
    client = _get_client(client_type)

    # Download the model
    client.download_model(
        model_name,
        bucket
    )


@airfiles.command('upload')
@click.argument('client-type', type=click.Choice(_ALLOWED_CLIENTS, case_sensitive=False))
@click.argument('model-path', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('--bucket', '-b', type=str, default=None)
@click.option('--overwrite/--no-overwrite', default=False)
def upload(client_type, model_path, bucket, overwrite):
    """
    Upload an .air file
    """

    # Create the client
    client = _get_client(client_type)

    # Upload the model
    client.upload_model(
        model_path,
        bucket,
        overwrite
    )



"""
The aisquared.config.harvesting subpackage contains objects for configuring harvesting of data.
"""

from .ImageHarvester import ImageHarvester
from .TextHarvester import TextHarvester
from .InputHarvester import InputHarvester



"""
The aisquared.serving package contains utilities to serve models to a local REST endpoint
"""

from mlflow.tensorflow import save_model as save_tensorflow_model
from mlflow.sklearn import save_model as save_sklearn_model
from mlflow.pytorch import save_model as save_pytorch_model
from mlflow.keras import save_model as save_keras_model

from .deploy_model import deploy_model, load_mann_model, load_keras_model, load_pytorch_model, load_sklearn_model, load_tensorflow_model
from .get_remote_prediction import get_remote_prediction



from aisquared.base import LOCATIONS, BaseObject


class ImageRendering(BaseObject):
    """
    Object which dictates how to render images
    """

    def __init__(
        self,
        color='blue',
        thickness='5px',
        placement=LOCATIONS[-1],
        include_probability=False,
        badge_color='white',
        font_color='black',
        font_size='5px',
        classes=None,
        threshold_key=None,
        threshold_value=None
    ):
        """
        Parameters
        ----------
        color : str (default 'blue')
            The color of the box around images
        thickness : str (default '5px')
            The thickness of the box around images
        placement : str (default 'bottomleft')
            The placement of the predicted value in the box
            around images
        include_probability : bool (default False)
            Whether to render calculated probabilities
        badge_color : str (default 'white')
            Background color of the text region
        font_color : str (default 'black')
            Color of the text
        font_size : str (default '5px')
            Text size
        classes : None or list (default None)
            If provided, list of classes that will be rendered
        threshold_key : None or str (default None)
            If provided, the key to use for thresholding
        threshold_value : None or numeric (default None)
            If provided with threshold_key, the value to use to threshold rendering
        """
        super().__init__()
        self.color = color
        self.thickness = thickness
        self.placement = placement
        self.include_probability = include_probability
        self.badge_color = badge_color
        self.font_color = font_color
        self.font_size = font_size
        self.classes = classes
        self.threshold_key = threshold_key
        self.threshold_value = threshold_value

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value

    @property
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, value):
        self._thickness = value

    @property
    def placement(self):
        return self._placement

    @placement.setter
    def placement(self, value):
        if value not in LOCATIONS:
            raise ValueError(f'placement must be one of {LOCATIONS}')
        self._placement = value

    @property
    def include_probability(self):
        return self._include_probability

    @include_probability.setter
    def include_probability(self, value):
        self._include_probability = value

    @property
    def badge_color(self):
        return self._badge_color

    @badge_color.setter
    def badge_color(self, value):
        self._badge_color = value

    @property
    def font_color(self):
        return self._font_color

    @font_color.setter
    def font_color(self, value):
        self._font_color = value

    @property
    def font_size(self):
        return self._font_size

    @font_size.setter
    def font_size(self, value):
        self._font_size = value

    @property
    def classes(self):
        return self._classes

    @classes.setter
    def classes(self, value):
        self._classes = value

    @property
    def threshold_key(self):
        return self._threshold_key

    @threshold_key.setter
    def threshold_key(self, value):
        self._threshold_key = value

    @property
    def threshold_value(self):
        return self._threshold_value

    @threshold_value.setter
    def threshold_value(self, value):
        self._threshold_value = value

    def to_dict(self):
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ImageRendering',
            'params': {
                'color': self.color,
                'thickness': self.thickness,
                'placement': self.placement,
                'includeProbability': self.include_probability,
                'badgeColor': self.badge_color,
                'fontColor': self.font_color,
                'fontSize': self.font_size,
                'classes': self.classes,
                'thresholdKey': self.threshold_key,
                'thresholdValue': self.threshold_value
            }
        }



"""
The aisquared.config subpackage contains utilities and objects for packaging aisquared configuration steps and models.
"""

from .ModelConfiguration import ModelConfiguration
from .GraphConfiguration import GraphConfiguration



from setuptools import setup
from aisquared import __version__

with open('requirements.txt', 'r') as f:
    requirements = [line for line in f.read().splitlines() if line != '']

setup(
    name='aisquared',
    version=__version__,
    url='https://github.com/AISquaredInc/aisquared',
    packages=[
        'aisquared',
        'aisquared.cli',
        'aisquared.base',
        'aisquared.remote',
        'aisquared.config',
        'aisquared.logging',
        'aisquared.serving',
        'aisquared.config.harvesting',
        'aisquared.config.preprocessing',
        'aisquared.config.preprocessing.text',
        'aisquared.config.preprocessing.image',
        'aisquared.config.preprocessing.tabular',
        'aisquared.config.analytic',
        'aisquared.config.postprocessing',
        'aisquared.config.rendering',
        'aisquared.config.feedback'
    ],
    author='The AI Squared Team',
    author_email='staff@squared.ai',
    description='Utilities for interacting with the AI Squared Technology Stack',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license='AI Squared Developer License Agreement',
    license_files='LICENSE',
    install_requires=requirements,
    classifiers=[
        'License :: Other/Proprietary License'
    ],
    entry_points={
        'console_scripts': [
            'aisquared = aisquared.cli.cli:aisquared'
        ]
    }
)



"""
The aisquared.remote package contains utilities for interacting with .air files stored in cloud
storage resources, specifically AWS S3 and Azure Blob Store
"""

from .aws import AWSClient
from .azure import AzureClient
